"use strict";
/*
Copyright 2021 The Matrix.org Foundation C.I.C.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getReleaseBefore = exports.getLatestRelease = exports.releasesContains = exports.getReleases = void 0;
const rest_1 = require("@octokit/rest");
// nb. this only gets a single pag, so if your're looking for a release
// more than 100 releases ago then, well, bad times.
function getReleases(owner, repo) {
    return __awaiter(this, void 0, void 0, function* () {
        const octo = new rest_1.Octokit();
        const rels = yield octo.rest.repos.listReleases({
            owner, repo, per_page: 100,
        });
        return rels.data;
    });
}
exports.getReleases = getReleases;
function releasesContains(rels, target) {
    return rels.some(r => r.name === target);
}
exports.releasesContains = releasesContains;
function getLatestRelease(rels, considerRCs) {
    if (considerRCs)
        return rels[0];
    for (const rel of rels) {
        if (!rel.prerelease)
            return rel;
    }
}
exports.getLatestRelease = getLatestRelease;
function getReleaseBefore(rels, target, considerRCs) {
    let found = false;
    for (const rel of rels) {
        if (rel.name === target) {
            found = true;
        }
        else if (found) {
            if (considerRCs || !rel.prerelease) {
                return rel;
            }
        }
    }
    if (!found) {
        throw new Error("Couldn't find release " + target);
    }
    else {
        throw new Error("Couldn't find release before " + target);
    }
}
exports.getReleaseBefore = getReleaseBefore;
